import functools
import json
import os
import urllib.request as req
import time
import random

import airline


def airline_wrapper(_handler=None, *, env_vars=None, add_role_info=True):
    '''airline decorator for a function in a AWS Batch job
    ```
    @airline_wrapper
    def my_handler(event, context):
        # ...
    ```
    '''

    def decorator_airline(handler):
        @functools.wraps(handler)
        def _airline_wrapper(*args, **kwargs):

            # don't blow up the world if the airline has not been initialized
            if not airline._ARL:
                return handler(*args, **kwargs)

            with airline._ARL.evented():
                airline.add_environment_variable('aws.batch.job_id', 'AWS_BATCH_JOB_ID')
                airline.add_environment_variable('aws.batch.compute_environment', 'AWS_BATCH_CE_NAME')
                airline.add_environment_variable('aws.batch.job_queue', 'AWS_BATCH_JQ_NAME')
                airline.add_environment_variable('aws.batch.job_attempt', 'AWS_BATCH_JOB_ATTEMPT')
                array_index = os.getenv('AWS_BATCH_JOB_ARRAY_INDEX')
                if array_index:
                    airline.add_context_field('aws.batch.array_job', True)
                    airline.add_context_field('aws.batch.array_index', array_index)

                if env_vars:
                    for name, var in env_vars.items():
                        airline.add_environment_variable(name, var)

                if add_role_info:
                    _add_role_info()

                resp = handler(*args, **kwargs)

                return resp
        return _airline_wrapper

    if _handler is None:
        return decorator_airline
    else:
        return decorator_airline(_handler)


def _add_role_info():
    uri = os.getenv("AWS_CONTAINER_CREDENTIALS_RELATIVE_URI")
    if not uri:
        return

    full_uri = f"http://169.254.170.2{uri}"
    data = _fetch_parse_uri(full_uri)

    airline.add_context_field('aws.role_arn', data.get('RoleArn'))
    airline.add_context_field('aws.access_key_id', data.get('AccessKeyId'))


def _fetch_parse_uri(uri, attempts=5):
    attempt = 0
    while True:
        try:
            attempt += 1
            with req.urlopen(uri, timeout=1) as f:
                response = json.load(f)
                return response
        except req.URLError:
            if attempt < attempts:
                _sleep(attempt)
            else:
                raise


def _sleep(attempt, cap=60, base=1):
    sleep_time = random.uniform(0, min(cap, base * 2**attempt))
    time.sleep(sleep_time)



import datetime as dt
from collections import defaultdict
from contextlib import contextmanager
import json
import time
from typing import (
    Dict,
    Any,
    Union,
    Optional,
    DefaultDict
)

from .format_exception import format_exception

Numeric = Union[int, float]
TIMER_PREFIX = 'timers.'
ROLLUP_PREFIX = 'rollup.'


class Event:
    def __init__(self, data: Dict[str, Any] = {}, created_at=dt.datetime.utcnow(), client=None):
        self._data: Dict[str, Any] = {}
        self._client = client

        if client:
            self.dataset = client.dataset
        else:
            self.dataset = ''
        self.created_at = created_at
        self.add(data=data)
        self._rollup_fields: DefaultDict[str, Numeric] = defaultdict(int)
        self._timer_fields: DefaultDict[str, float] = defaultdict(float)

    def add(self, data: Dict[str, Any]):
        self._data.update(data)

    def add_field(self, name: str, value: Any):
        self._data[name] = value

    def add_rollup_field(self, name: str, value: Numeric):
        self._rollup_fields[name] += value

    @contextmanager
    def add_timer_field(self, name: str):
        try:
            start = time.perf_counter()
            yield
        finally:
            done = time.perf_counter()
            self._timer_fields[name] += (done - start) * 1000

    def attach_exception(self, err: Optional[BaseException] = None, prefix: str = 'exception'):
        self.add(format_exception(err, prefix))

    def __str__(self):
        return json.dumps({
            "data": self._data,
            "rollups": self._rollup_fields,
            "timers": self._timer_fields
        })

    def send(self):
        if self._client:
            self._client.send(self)
        else:
            raise RuntimeError("Can't send, no client!")

    def rollup_fields(self) -> Dict[str, Numeric]:
        return {_rollup_name(k): v for k, v in self._rollup_fields.items()}

    def fields(self) -> Dict[str, Any]:
        return {**self._data, **self.rollup_fields(), **self.timer_fields()}

    def timer_fields(self) -> Dict[str, float]:
        return {_timer_name(k): round(v, 3) for k, v in self._timer_fields.items()}


def _rollup_name(name: str):
    if name.startswith(ROLLUP_PREFIX):
        return name
    else:
        return ROLLUP_PREFIX + name


def _timer_name(name: str):
    if not name.startswith(TIMER_PREFIX):
        name = TIMER_PREFIX + name

    if not name.endswith('_ms'):
        name = name + '_ms'
    return name



import airline.client as sl
import pytest


def test_dots_to_deep_not_deep():
    assert sl.dots_to_deep({'a': 1}) == {'a': 1}


def test_dots_to_deep_one_level():
    assert sl.dots_to_deep({'a.b': 1}) == {'a': {'b': 1}}


def test_dots_to_deep_two_level():
    assert sl.dots_to_deep({'a.b': 1, 'c': 2}) == {'a': {'b': 1}, 'c': 2}


def test_dots_to_deep_three_level():
    assert sl.dots_to_deep({'a.b.c': 1, 'c': 2}) == {'a': {'b': {'c': 1}}, 'c': 2}


def test_dots_to_deep_warns_on_collision():
    with pytest.raises(RuntimeError):
        sl.dots_to_deep({'a': 1, 'a.b': 1, 'c': 2})


def test_dots_to_deep_warns_on_collision_no_matter_the_order():
    with pytest.raises(RuntimeError):
        sl.dots_to_deep({'a.b': 1, 'a': 1, 'c': 2})


def test_dots_to_deep_warns_on_deep_collisions():
    with pytest.raises(RuntimeError):
        sl.dots_to_deep({'a.b.c': 1, 'a.b': 1})


def test_dots_to_deep_warns_on_deep_collisions2():
    with pytest.raises(RuntimeError):
        sl.dots_to_deep({'a.b': 1, 'a.b.c': 1, })


def test_dots_to_deep_warns_on_deep_collisions3():
    with pytest.raises(RuntimeError):
        sl.dots_to_deep({'a': {'b': 1}, 'a.b.c': 1, })


def test_dots_to_deep_warns_on_deep_collisions4():
    with pytest.raises(RuntimeError):
        sl.dots_to_deep({'a': {'b': 1}, 'a.b.c': 1, })



from contextlib import contextmanager
import json
import logging
import time
import sys
from typing import (
    Optional,
    Dict,
    Any,
    Union,
)

from .event import Event
from .version import __version__


log = logging.getLogger('airline')


class Client():
    def __init__(self, dataset: str, debug=False):
        self.dataset = dataset
        self.debug = debug

    def add_context_field(self, name: str, value: Any):
        if self._event:
            self._event.add_field(name, value)
        else:
            self.log("No event found")

    def add_context(self, data: Dict[str, Any]):
        if self._event:
            self._event.add(data=data)
        else:
            self.log("No event found")

    def add_rollup_field(self, name: str, value: Union[int, float]):
        if self._event:
            self._event.add_rollup_field(name, value)
        else:
            self.log("No event found")

    @contextmanager
    def add_timer_field(self, name: str):
        if self._event:
            with self._event.add_timer_field(name):
                yield
        else:
            self.log("No event found")
            yield

    def attach_exception(self, err: Optional[BaseException] = None, prefix: str = 'exception'):
        if self._event:
            self._event.attach_exception(err, prefix)
        else:
            self.log("No event found")

    @contextmanager
    def evented(self):
        if self._event:
            self.log("Event already created")
        else:
            event = self.start()
        start = time.perf_counter()
        try:
            yield
        except Exception as e:
            event.add_field('status', 'ERROR')
            event.attach_exception(e)
            raise
        finally:
            done = time.perf_counter()
            duration = (done - start) * 1000
            event.add_field('duration_ms', round(duration, 3))
            self.done()

    def start(self, event=None):
        if event is None:
            event = self.new_event()

        self._event = event
        return event

    def done(self):
        self.send(self._event)
        self._event = None

    def new_event(self, data={}):
        return Event(data=data, client=self)

    def send(self, ev: Event):
        '''send accepts an event and writes it to the configured output file'''
        event_time = ev.created_at.isoformat()
        if ev.created_at.tzinfo is None:
            event_time += "Z"

        payload = {
            "time": event_time,
            "dataset": ev.dataset,
            "client": "airline/" + __version__,
            "data": dots_to_deep(ev.fields()),
        }
        if self.debug:
            indent: Optional[int] = 2
        else:
            indent = None

        print(json.dumps(payload, indent=indent, default=_json_default_handler) + "\n", file=sys.stderr)

    def log(self, message, *args):
        if self.debug:
            log.debug(message, *args)

    def __repr__(self):
        return "{cls}(dataset={dataset!r}, debug={debug!r})".format(
            cls=self.__class__.__name__,
            dataset=self.dataset,
            debug=self.debug,
        )


def _json_default_handler(obj):
    if isinstance(obj, Event):
        return obj.fields()
    try:
        return str(obj)
    except TypeError:
        return repr(obj)


def dots_to_deep(dictionary):
    new_dict = {}
    for k, v in dictionary.items():
        *keys, key = k.split('.')
        d = new_dict
        for kk in keys:
            if kk not in d:
                d[kk] = {}
            d = d[kk]
        if not isinstance(d, dict):
            raise RuntimeError("Incorrect nesting specified: key=%s collides with key=%s" % (k, '.'.join(keys)))

        if key in d:
            raise RuntimeError("Incorrect nesting specified: key=%s would overwrite nesting key" % (k,))

        d[key] = v

    return new_dict



import threading

from .client import Client


class ThreadLocalClient(Client):
    def __init__(self, *args, **kwargs):
        super(ThreadLocalClient, self).__init__(*args, **kwargs)
        self._state = threading.local()

    @property
    def _event(self):
        return getattr(self._state, 'event', None)

    @_event.setter
    def _event(self, new_event):
        self._state.event = new_event



"""
Lightweight wide event logging to bring more observability to lambda functions.

The idea is to build up the full context of a function/script into one wide
event that gets emitted at the end.  This puts all the context in one log
message, making it very easy to run analytics on, find like errors, and a lot
of other things.  A full blown observability platform like honeycomb would be
more informative, and allows for the notion of spans and distributed tracing
(i.e. across different (micro) services and the like).

But this is a start.
"""
from contextlib import contextmanager
import functools
import logging
import os
from typing import (
    Dict,
    Any,
    Optional,
)

from .threadlocal_client import ThreadLocalClient
from .version import __version__   # noqa: F401


_ARL = None


log = logging.getLogger('airline')


def init(dataset: str = '', debug=False):
    global _ARL

    if _ARL is None:
        _ARL = ThreadLocalClient(dataset=dataset, debug=debug)
    else:
        log.warning("Library already initialized: client=%r new_dataset=%s", _ARL, dataset)


def add_context(data: Dict[str, Any]):
    '''Similar to add_context_field(), but allows you to add a number of name:value pairs
    to the currently active event at the same time.
    `airline.add_context({ "first_field": "a", "second_field": "b"})`
    Args:
    - `data`: dictionary of field names (strings) to field values to add
    '''
    if _ARL:
        _ARL.add_context(data=data)


def add_context_field(name: str, value: Any):
    ''' Add a field to the currently active event. For example, if you are
    using django and wish to add additional context to the current request
    before it is sent:
    `airline.add_context_field("my field", "my value")`
    Args:
    - `name`: Name of field to add
    - `value`: Value of new field
    '''
    if _ARL:
        _ARL.add_context_field(name=name, value=value)


def add_rollup_field(name: str, value: Any):
    ''' AddRollupField adds a key/value pair to the current event. If it is called repeatedly
    on the same event, the values will be summed together.
    Args:
    - `name`: Name of field to add
    - `value`: Numeric (float) value of new field
    '''

    if _ARL:
        _ARL.add_rollup_field(name=name, value=value)


def add_environment_variable(name: str, env_var: str):
    """
    AddEnvironmentVariable adds the value of an environment variable to the current event.
    Args:
    - `name`: Name of the field to add
    - `env_var`: Name of the environment variable
    """

    if _ARL:
        _ARL.add_context_field(name=name, value=os.getenv(env_var))


@contextmanager
def timer(name: str):
    """ Timer yields block (think `with` statement) and counts the time
     taken during that block.  The time is added to the event.  If there
     are multiple invocations with the same name, these will be added up
     over the whole event.
    It is especially useful for doing things like adding the duration spent talking
    to a specific external service - eg database time
    """
    if _ARL:
        with _ARL.add_timer_field(name):
            yield
    else:
        yield


def done():
    ''' close the airline client, flushing any unsent events. '''
    global _ARL
    if _ARL:
        _ARL.done()

    _ARL = None


def evented(**extra_context):
    """Decorator for wrapping a generic function in an event.

    The event will be sent when the function ends, possibly annotated with
    any exception raised."""
    def wrapped(fn):
        @functools.wraps(fn)
        def inner(*args, **kwargs):
            if _ARL:
                with _ARL.evented():
                    if extra_context:
                        add_context(extra_context)
                    return fn(*args, **kwargs)
            else:
                return fn(*args, **kwargs)

        return inner

    return wrapped


def timed(add_count=True, **extra_context):
    """
    Decorator to turn a function call into a timed function call.
    The total duration of all calls to the function, and the number
    of calls will be attached as fields to the event.
    Also, additional context can be provided.
    """
    def wrapped(fn):
        @functools.wraps(fn)
        def inner(*args, **kwargs):
            if _ARL:
                name = fn.__name__
                timer_name = f"{name}_duration"
                count_name = f"{name}_calls"
                with _ARL.add_timer_field(timer_name):
                    if add_count:
                        _ARL.add_rollup_field(count_name, 1)
                    if extra_context:
                        add_context(extra_context)
                    return fn(*args, **kwargs)
            else:
                return fn(*args, **kwargs)

        return inner

    return wrapped


def attach_exception(err: Optional[BaseException] = None, prefix: str = 'exception'):
    """
    Attach an exception and traceback to the current event with the given prefix
    """
    if _ARL:
        _ARL.attach_exception(err=err, prefix=prefix)


def set_status(status: str):
    """
    Set the status field on the current event
    """
    add_context_field('status', status)


def success():
    """
    helper methods for consistent success/fail statuses
    """
    set_status('SUCCESS')


def error():
    """
    helper methods for consistent success/fail statuses
    """
    set_status('ERROR')



import functools

import airline


COLD_START = True


def airline_wrapper(_handler=None, *, add_event=False, add_response=False):
    '''airline decorator for Lambda functions. Expects a handler
    function with the signature:
    `def handler(event, context)`
    Example use:
    ```
    @airline_wrapper
    def my_handler(event, context):
        # ...
    ```
    '''

    def decorator_airline(handler):
        @functools.wraps(handler)
        def _airline_wrapper(event, context):
            global COLD_START

            # don't blow up the world if the airline has not been initialized
            if not airline._ARL:
                return handler(event, context)

            try:

                with airline._ARL.evented():
                    airline.add_context({
                        "app.function_name": getattr(context, 'function_name', ""),
                        "app.function_version": getattr(context, 'function_version', ""),
                        "app.request_id": getattr(context, 'aws_request_id', ""),
                        "meta.cold_start": COLD_START,
                    })

                    if add_event:
                        airline.add_context_field("app.event", event)

                    resp = handler(event, context)

                    if resp is not None and add_response:
                        airline.add_context_field('app.response', resp)

                    return resp
            finally:
                # This remains false for the lifetime of the module
                COLD_START = False

        return _airline_wrapper

    if _handler is None:
        return decorator_airline
    else:
        return decorator_airline(_handler)



import pytest

import airline.client


@pytest.fixture()
def client() -> airline.client.Client:
    client = airline.client.Client('test')
    return client


def test_dataset_is_available(client):
    assert client.dataset == 'test'


def test_a_client_can_be_started(client):
    client.start()

    assert client._event is not None


def test_add_context_field_adds_to_event(client, mocker):
    ev = mocker.MagicMock()
    client.start(ev)

    client.add_context_field('foo', 'bar')

    ev.add_field.assert_called_with('foo', 'bar')


def test_add_context_adds_to_event(client, mocker):
    ev = mocker.MagicMock()
    client.start(ev)

    client.add_context({'foo': 'bar'})

    ev.add.assert_called_with(data={'foo': 'bar'})


def test_add_rollup_field_adds_to_event(client, mocker):
    ev = mocker.MagicMock()
    client.start(ev)

    client.add_rollup_field('foo', 1)

    ev.add_rollup_field.assert_called_with('foo', 1)



import sys
import traceback
from typing import (
    Optional
)

FS_NAMES = ('filename', 'lineno', 'name',)


def format_exception(exception: Optional[BaseException] = None, prefix: str = 'exception') -> dict:
    cls, msg, tb = _extract_exception(exception)

    return {
        f'{prefix}.type': cls.__name__,
        f'{prefix}.message': str(msg),
        f'{prefix}.traceback': [_fs_to_dict(i) for i in traceback.extract_tb(tb)],
    }


def _fs_to_dict(fs: traceback.FrameSummary, fs_names=FS_NAMES):
    return {n: getattr(fs, n) for n in fs_names}


def _extract_exception(exception: Optional[BaseException] = None):
    if isinstance(exception, BaseException):
        return exception.__class__, exception, exception.__traceback__
    elif isinstance(exception, tuple):
        return exception
    else:
        return sys.exc_info()



import airline.format_exception


def test_format_exception_works():
    try:
        raise RuntimeError('example')
    except RuntimeError as e:
        v = airline.format_exception.format_exception(e)

        assert v['exception.type'] == "RuntimeError"
        assert v['exception.message'] == 'example'
        assert len(v['exception.traceback']) == 1

        trace = v['exception.traceback']
        first = trace[0]

        assert 'filename' in first
        assert first['lineno'] == 6
        assert first['name'] == 'test_format_exception_works'


def test_format_exception_works_with_no_explicit_exception():
    try:
        raise RuntimeError('example')
    except RuntimeError:
        v = airline.format_exception.format_exception()

        assert v['exception.type'] == "RuntimeError"
        assert v['exception.message'] == 'example'
        assert len(v['exception.traceback']) == 1

        trace = v['exception.traceback']
        first = trace[0]

        assert 'filename' in first
        assert first['lineno'] == 24
        assert first['name'] == 'test_format_exception_works_with_no_explicit_exception'


def test_format_exception_works_with_prefix():
    try:
        raise RuntimeError('example')
    except RuntimeError as e:
        v = airline.format_exception.format_exception(e, prefix='foo')

        assert v['foo.type'] == "RuntimeError"
        assert v['foo.message'] == 'example'
        assert len(v['foo.traceback']) == 1

        trace = v['foo.traceback']
        first = trace[0]

        assert 'filename' in first
        assert first['lineno'] == 42
        assert first['name'] == 'test_format_exception_works_with_prefix'



import pytest

import airline.event


@pytest.fixture
def event():
    return airline.event.Event()


def test_an_added_field_is_in_fields(event):
    event.add_field('foo', 'bar')

    assert event.fields()['foo'] == 'bar'


def test_added_fields_are_in_fields(event):
    event.add({'foo': 'bar', 'baz': 1})

    assert event.fields() == {'foo': 'bar', 'baz': 1}


@pytest.mark.parametrize('name', ['foo', 'rollup.foo'])
def test_rollup_fields_are_added(event, name):
    event.add_rollup_field(name, 1)

    assert event.fields() == {'rollup.foo': 1}


def test_rollup_fields_accumulate(event):
    event.add_rollup_field('foo', 1)
    event.add_rollup_field('foo', 1)

    assert event.fields() == {'rollup.foo': 2}


def test_timers_work(event, mocker):
    mocker.patch('time.perf_counter', side_effect=[0, 5])

    with event.add_timer_field('five'):
        pass

    assert event.fields() == {'timers.five_ms': 5000.0}


@pytest.mark.parametrize('name', ['example', 'timers.example', 'example_ms', 'timers.example_ms'])
def test_timer_fields_are_added(event, name):
    with event.add_timer_field(name):
        pass

    assert 'timers.example_ms' in event.fields()


def test_fields_are_in___str__(event):
    event.add_field('foo', 1)

    assert '"data": {"foo": 1' in str(event)


def test_rollups_are_in___str__(event):
    event.add_rollup_field('foo', 1)

    assert '"rollups": {"foo": 1' in str(event)


def test_timers_are_in___str__(event):
    with event.add_timer_field('foo'):
        pass

    assert '"timers": {"foo": ' in str(event)


def test_send_passes_off_to_client(mocker):
    client = mocker.Mock()

    event = airline.event.Event(client=client)

    event.send()

    client.send.assert_called_with(event)


def test_attach_exception_works(event):
    try:
        raise RuntimeError('example')
    except RuntimeError as e:
        event.attach_exception(e)

    assert 'exception.type' in event.fields()
    assert 'exception.message' in event.fields()
    assert 'exception.traceback' in event.fields()


def test_attach_exception_works2(event):
    try:
        raise RuntimeError('example')
    except RuntimeError as e:
        event.attach_exception(e, prefix='bar')

    assert 'bar.type' in event.fields()
    assert 'bar.message' in event.fields()
    assert 'bar.traceback' in event.fields()



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['airline']

package_data = \
{'': ['*']}

extras_require = \
{'test': ['pytest ~=5.3', 'pytest-mock ~=2.0']}

setup(name='airline',
      version='0.5.1',
      description='Lightweight wide event logging to bring more observability to lambda functions.',
      author='namelessjon',
      author_email='jonathan.stott@gmail.com',
      url='https://github.com/namelessjon/airline',
      packages=packages,
      package_data=package_data,
      extras_require=extras_require,
     )
